(* operators *)
type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq | Exp | Concat | And | Or
type unop = Neg | Not

(* expressions *)
type expr =
    Literal of literal
  | Id of string
  | Binop of expr * op * expr
  | Assign of string * expr
  | Unop of unop * expr
  | Call of string * expr list
  | Li_list of expr list
  | Li_list_ref of string * expr
  | Noexpr

(* literal *)
and literal =
  | Int_Literal of int
  | Double_Literal of float
  | String_Literal of string
  | Bool_Literal of bool
  | Li_Literal of expr list

type var_type =
  | Int
  | String
  | Bool
  | Double
  | Li of var_type
  | None 

(* variable declarations *)
type var_decl =
  | Var of var_type * string                            (* int a *)
  | Var_Decl_Assign of var_type * string * expr         (* double b = 1.1 *)
  
type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | Loop of expr * expr * expr * stmt
  | Var_decl of var_decl
 
type formal = 
  | Formal of var_type * string

(* function declaration *)
type func_decl = {
    fname : string;
    formals : formal list;
    body : stmt list;
    return_type: var_type;
  }

type program = var_decl list * func_decl list

(* get string version of all types *)
let string_of_binop = function
  | Add          -> "+"
  | Sub          -> "-"
  | Mult         -> "*"
  | Div          -> "/"
  | Equal        -> "=="
  | Neq          -> "!="
  | Less         -> "<"
  | Leq          -> "<="
  | Greater      -> ">"
  | Geq          -> ">="
  | And          -> "&&"
  | Or           -> "||"
  | Concat       -> "~"
  | Exp          -> "^"

let string_of_unop = function
  | Neg -> "-"
  | Not -> "!"

let rec string_of_expr = function
    Literal(l) -> string_of_literal l
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      string_of_binop o ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Unop(o, e) ->
      string_of_unop o ^ " " ^
      string_of_expr e
  | Li_list(li) -> "[" ^ String.concat ", " (List.map string_of_expr li) ^ "]"
  | Li_list_ref(li, ind) -> li ^ "[" ^ string_of_expr ind ^ "]"
  | Noexpr -> ""

and string_of_literal = function
  | Int_Literal(l)    -> string_of_int l
  | Double_Literal(l) -> string_of_float l
  | String_Literal(l) -> "\"" ^ l ^ "\""
  | Bool_Literal(l)   -> string_of_bool l
  | Li_Literal(l)    -> "[" ^ String.concat ", " (List.map string_of_expr l) ^ "]"


let rec string_of_var_type = function
  | Int -> "int"
  | Double -> "double"
  | String -> "string"
  | Li(t) -> string_of_var_type t ^ " list"
  | Bool -> "bool"
  | None -> "none"

let string_of_formal = function
  | Formal(t, id) -> string_of_var_type t ^ " " ^ id ^ ";"


let string_of_vdecl = function 
  | Var(t, id) -> string_of_var_type t ^ " " ^ id ^ ";"
  | Var_Decl_Assign(t, id, expr) -> string_of_var_type t ^ " " ^ id ^ "=" ^ string_of_expr expr ^ ";"

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | Loop(v1, e, v2, s) -> "loop conditions ( start: " ^ string_of_expr v1 ^ "; check: " ^ string_of_expr e ^ "; change: " ^ string_of_expr v2 ^ ") do " ^ string_of_stmt s
  | Var_decl(v) -> string_of_vdecl v
 

let string_of_fdecl fdecl =
  "function " ^ fdecl.fname ^ " returns "^ string_of_var_type fdecl.return_type ^ " (" ^ String.concat ", " (List.map string_of_formal fdecl.formals) ^ ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
