open Ast
open Sast
open Printf

let imports =
  "$( document ).ready(function() {\n\t"

let set_default_val ty = match ty with
  Int -> "0"
  | String -> "\"\""
  | Double -> "0.0"
  | Bool -> "false"
  | Li(t) -> "[]"
  | _ -> "\"\""


let rec gen_var_type = function _ -> "var"

let gen_binop = function
  | Add          -> "+"
  | Sub          -> "-"
  | Mult         -> "*"
  | Div          -> "/"
  | Equal        -> "=="
  | Neq          -> "!="
  | Less         -> "<"
  | Leq          -> "<="
  | Greater      -> ">"
  | Geq          -> ">="
  | And          -> "&&"
  | Or           -> "||"
  (* see below for actual concat/exp use *)
  | Concat       -> "~"
  | Exp          -> "^"
  
 let gen_unop = function
  | Not -> "!"
  | Neg -> "-"

let gen_literal lit = match lit with
  | SInt_Literal(i)      -> string_of_int i
  | SDouble_Literal(d)   -> string_of_float d
  | SBool_Literal(b)     -> string_of_bool b
  | SString_Literal(str) -> "\"" ^ str ^ "\""

let rec gen_literal_list ll = match ll with
  | [] -> ""
  | head::[] -> gen_literal head
  | head::tail -> gen_literal head ^ ", " ^ gen_literal_list tail

let rec gen_expr expr = match expr with
  | SLiteral(l,t)           -> gen_literal l
  | SId(v,t)                -> v
  | SUnop(u, e, t)          -> gen_unop u ^ "(" ^ gen_expr e ^ ")"
  | SBinop(e1, op, e2, t)   -> (match op with
            | Concat -> gen_expr e1 ^ ".concat(" ^ gen_expr e2 ^ ")"
            | Exp    -> "Math.pow(" ^ gen_expr e1 ^ ", " ^ gen_expr e2 ^ ")"
            | _ ->gen_expr e1 ^ gen_binop op ^ gen_expr e2)
  | SCall(id, e1, t)        -> if(id="print") then gen_print e1
                                else if(id="rand") then "Math.floor((Math.random() * " ^ gen_expr (List.hd e1) ^" + 1))"
                                else if(id="str") then gen_expr (List.hd e1) ^ ".toString()"
                                else if(id="evalDouble") then "eval(" ^ gen_expr (List.hd e1) ^ ")"
                                else if(id="evalInt") then "Math.floor(eval(" ^ gen_expr (List.hd e1) ^ "))"
                                else if (id="get_char_at") then gen_expr(List.hd e1)^".charAt("^gen_expr(List.nth e1 1)^")"
                                else if (id="length") then gen_expr (List.hd e1) ^".length"
                                else if (id="strReplace") then gen_expr(List.hd e1)^".replace("^gen_expr(List.nth e1 1)^","^gen_expr(List.nth e1 2)^")"
                                else id ^ "(" ^ gen_expr_list e1 ^ ")"
  | SAssign(str, e1, t)     -> str ^ " = " ^ gen_expr e1
  | SLi_list(li, t)           -> "[" ^ String.concat "," (List.map gen_expr li) ^ "]"
  | SLi_list_ref(str, i, t)   -> str ^ "[" ^ string_of_expr i ^ "]"
  | SNoexpr                  -> "\"\""

and gen_expr_list expr_list = match expr_list with
  | [] -> ""
  | head::[] -> gen_expr head
  | head::tail -> gen_expr head ^ ", " ^ gen_expr_list tail

and gen_formal h = match h with
  SFormal(type_spec, id) -> id

and gen_formal_list fl = match fl with
  | [] -> ""
  | head::[] -> gen_formal head
  | head::tail -> gen_formal head ^ ", " ^ gen_formal_list tail

and gen_print p = match p with
  | []         -> ""
  | head::[]   -> "console.log(" ^ gen_expr head ^ ")"
  | head::tail -> "console.log(" ^ gen_expr head ^ ")" ^ gen_print tail


and gen_sstmt stmt = match stmt with
  | SBlock(stmt_list) -> gen_sstmt_list stmt_list
  | SExpr(expr)       -> gen_expr expr ^ ";\n\t"
  | SReturn(expr) -> "\treturn " ^ gen_expr expr ^ ";\n\t"
  | SIf(expr,s1,s2) -> "\tif(" ^ gen_expr expr ^ ") {\n\t\t" ^ gen_sstmt s1 ^ "}\n\t\telse {\n\t" ^ gen_sstmt s2 ^ "}\n\n"
  | SLoop(e1, e2, e3, s1) -> "\tfor (" ^ gen_expr e1 ^ "; " ^ gen_expr e2 ^ "; " ^ gen_expr e3 ^ "){\n\t\t" ^ gen_sstmt s1 ^ "}\n"
  | SVar_Decl(vdec) -> gen_var_dec vdec ^"\n\t"

and gen_sstmt_list stmt_list = match stmt_list with
  | [] -> ""
  | head::[] -> gen_sstmt head
  | head::tail -> gen_sstmt head ^ gen_sstmt_list tail

and gen_var_dec dec = match dec with
  | SVar(ty,id) -> gen_var_type ty ^ " " ^ id ^ ";\n\t"
  | SVar_Decl_Assign(ty,id,e) -> "var " ^ id ^ " = " ^ gen_expr e ^ ";\n\t"
  
and gen_var_dec_list var_dec_list = match var_dec_list with
  | [] -> ""
  | head::[] -> gen_var_dec head
  | head::tail -> gen_var_dec head ^ gen_var_dec_list tail

and gen_global_var_dec_list var_dec_list = match var_dec_list with
  | [] -> ""
  | head::[] -> gen_var_dec head
  | head::tail -> gen_var_dec head ^ gen_var_dec_list tail

and gen_func_dec func =
  if(func.sfname = "run") then gen_sstmt_list func.sbody
  else "\tfunction " ^ func.sfname ^ "(" ^ gen_formal_list func.sformals ^ ") {\n\t" ^ gen_sstmt_list func.sbody ^ "}\n"

and gen_func_dec_list fl = match fl with
  | [] -> ""
  | head::[] -> gen_func_dec head
  | head::tail -> gen_func_dec head ^ gen_func_dec_list tail

let prog_gen = function
  SProg(checked_globals, checked_functions) ->
    imports ^
    gen_global_var_dec_list (List.rev checked_globals) ^ "\n" ^
    gen_func_dec_list checked_functions ^
    "\n});\n"

