%{
   open Ast
%}

%token LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET
%token SEMI COMMA CONCAT
%token LOOP START CONDITIONS CHECK CHANGE DO COLON
%token PLUS MINUS TIMES DIVIDE EXP ASSIGN 
%token EQ NEQ LT LEQ GT GEQ AND OR NOT
%token IF ELSE
%token FUNCTION RETURN RETURNS
%token INT DOUBLE STRING BOOL LI NONE
%token <int> INT_LITERAL
%token <string> STRING_LITERAL ID
%token <float> DOUBLE_LITERAL
%token <bool> BOOL_LITERAL
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left CONCAT
%left PLUS MINUS
%left TIMES DIVIDE
%left EXP
%left NEG NOT

%start program
%type <Ast.program> program

%%

program:
  | decls EOF         { $1 }

decls:
  | /* nothing */                  { ([], []) }
  | decls global_var_declaration { ($2 :: fst $1), snd $1 }
  | decls function_declaration   { fst $1, ($2 :: snd $1) }

var_declaration:
  | var_type ID SEMI                  { Var($1, $2) }                     /* int a; */
  | var_type ID ASSIGN expr SEMI      { Var_Decl_Assign($1, $2, $4) }     /* int a = 1; */
  | var_type LI ID ASSIGN expr SEMI  { Var_Decl_Assign($1, $3, $5) }      /* int list a = [1,2,3] */

global_var_declaration:
  var_declaration { $1 }

var_type:
  | INT                 { Int }
  | STRING              { String }
  | DOUBLE              { Double }
  | BOOL                { Bool }
/*  | var_type LI         { Li($1) } */
  | NONE                { None }


function_declaration:
 /* function foo return int (int a, int b) { ... } */
 | FUNCTION ID RETURNS var_type LPAREN formal_parameters RPAREN LBRACE statements RBRACE
     {
       { fname = $2;
         formals = $6;
         return_type = $4;
         body = List.rev $9;
       }
     }

formal_parameters:
  | /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
  | parameter                   { [$1] }     /* foo: Int */
  | formal_list COMMA parameter { $3 :: $1 } /* foo: Int, bar: String */

parameter:
  | var_type ID    { Formal($1, $2) }        /* foo: Int */

statements:
  | /* nothing */        { [] }
  | statements statement { $2 :: $1 }

statement:
  | expr SEMI                                                 { Expr($1) }                  /* 1 + 2 */
  | RETURN expr SEMI                                          { Return($2) }                /* return 1 + 2 */
  | LBRACE statements RBRACE                                  { Block(List.rev $2) }        /* { 1 + 2 \n 3 + 4 } */
  | IF LPAREN expr RPAREN statement %prec NOELSE              { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN statement ELSE statement            { If($3, $5, $7) }
  | LOOP CONDITIONS LPAREN START COLON assign_opt SEMI
   CHECK COLON expr_comp SEMI
   CHANGE COLON expr_opt RPAREN DO statement                  { Loop($6, $10, $14, $17) }   /* loop coditions (start: a=1; check: a<5; a=a+1) { ... } */
  | var_declaration                                           { Var_decl($1) }              /* see above */
 


expr_opt:
  | /*nothing*/    { Noexpr } 
  | expr           { $1 }

expr:
  | literal                      { Literal($1) }                /* 18, "Awesome", 3.14, true */
  | binary_operation             { $1 }                         /* 4 + 3, "Two " ~ "words" */
  | unary_operation              { $1 }                         /* -1 */
  | ID                           { Id($1) }                     /* question, answer, option */
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }               /* display_radio(question, answer, name) */
  | LPAREN expr RPAREN           { $2 }                         /* (4 + 6) */
  | assign                       { $1 }                         /* a = 1 */
  | LBRACKET list_opt RBRACKET   { Li_list($2) }                /* [1,2,3,4] */
  | ID LBRACKET expr RBRACKET    { Li_list_ref($1, $3)}         /* answer[0] */
         


list_opt:
  | /*nothing*/   { [] }
  | li_list       { List.rev $1 }

li_list:
  | expr                  { [$1] }
  | li_list COMMA expr    { $3 :: $1 }

assign_opt:
  | /*nothing*/    { Noexpr } 
  | assign           { $1 }

  assign:
  | ID ASSIGN expr              { Assign($1, $3)}

literal:
  | INT_LITERAL                  { Int_Literal($1) }          /* 1, 7, 100 */
  | STRING_LITERAL               { String_Literal($1) }       /* "Marissa", "Michelle", "Carolyn" */
  | DOUBLE_LITERAL               { Double_Literal($1) }       /* 2.78 */
  | BOOL_LITERAL                 { Bool_Literal($1) }         /* true, false */

actuals_opt:
  | /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
  | expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }

binary_operation:
  | expr PLUS   expr        { Binop($1, Add,   $3) }
  | expr MINUS  expr        { Binop($1, Sub,   $3) }
  | expr TIMES  expr        { Binop($1, Mult,  $3) }
  | expr DIVIDE expr        { Binop($1, Div,   $3) }
  | expr CONCAT expr        { Binop($1, Concat, $3) }
  | expr EXP    expr        { Binop($1, Exp, $3)}
  | expr_comp               { $1 }

expr_comp:
  | expr EQ     expr        { Binop($1, Equal, $3) }
  | expr NEQ    expr        { Binop($1, Neq,   $3) }
  | expr LT     expr        { Binop($1, Less,  $3) }
  | expr LEQ    expr        { Binop($1, Leq,   $3) }
  | expr GT     expr        { Binop($1, Greater,  $3) }
  | expr GEQ    expr        { Binop($1, Geq,   $3) }
  | expr AND    expr        { Binop($1, And, $3) }
  | expr OR     expr        { Binop($1, Or, $3) }

unary_operation:
  | NOT expr                     { Unop(Not, $2) }
  | MINUS expr %prec NEG         { Unop(Neg, $2) }